/****************************************************************************
 *
 *   File: MLGraph.h:
 *   
 *   Author: Mark Stephenson
 *   Date: Mon Dec  1 13:59:26 2003
 *
 *   Function: A dense matrix graph representation.  The ORC graph is hard
 *      to query and modify, so I'm making my own.
 *
 ***************************************************************************/

#ifndef __MLGRAPH_H__
#define __MLGRAPH_H__

#include "cg_dep_graph.h"
#include "cg_loop_scc.h"
#include "op_list.h"
#include "whirl2ops.h"
#include "findloops.h"
#include "op.h"
#include "bb.h"
#include <vector>
#include <list>
#include <set>


typedef struct {
  ARC   *edge;
  UINT8  omega;
} MLEdge;


#define lcdmin(var, omega) \
if ((omega) && ((omega) < (var))) \
  (var) = (omega);

class MLGraph 
{
 public:

  MLGraph (BB *loophead, bool cyclic);
  ~MLGraph ();

  void print (INT, char*);

  void find_sccs ();
  void degrees (double& in_degree, 
		double& out_degree);

  void dep_height (UINT32& num_sccs, 
		   UINT32& max_dh, 
		   UINT16& max_memdep,
		   UINT16& cont_dep,
 		   double& avg_dh, 
		   UINT32& non_trivial);

  void cyclic_sccs (UINT32& num_sccs,
		    UINT32& non_trivial);

  void find_indirect_refs (UINT32&);
  void minimum_lcd (UINT32&, UINT32&, UINT32&,
		    UINT32&, UINT32&, UINT32&);

 protected:

  int find_sccs (int*, int*, int);
  void add_edge (ARC*);
  UINT32 get_address (UINT32 m, UINT32 n);
  UINT32 find (OP*);

  BOOL induction_var (OP*, INT);
  BOOL normal_edge (const MLEdge& edge)
    {
      ARC* val = edge.edge;
      return ((val != NULL) && (val != m_hallucinate));
    }

  BOOL normal_edge_present (INT k, INT t)
    {
      return normal_edge(m_edges[get_address(k, t)]);
    }

  BOOL edge_present (INT k, INT t)
    {
      return (m_edges[get_address(k, t)].edge != NULL);
    }

  BOOL floating_store (OP*);

  void depth (UINT32, UINT32&, UINT16&);
  void max_depth (set<UINT32>*, UINT32&, UINT16&, UINT16&);

  OP  *single_succ (OP*);
  void find_invariants ();
  BOOL connected_to_load (INT);

 protected:

  BOOL m_cyclic;
  BB *m_loophead;
  vector<OP*> m_nodes;
  list<set<UINT32>*> m_sccs;
  MLEdge *m_edges;
  ARC *m_hallucinate;
  UINT64 *m_height;
  UINT32 m_p;
  UINT32 m_id;
  UINT32 m_arith_store;
  UINT32 m_load_arith_store;
  UINT32 m_invariants;
};



#endif 
