/****************************************************************************
 *
 *   File: timers.h:
 *   
 *   Author: Mark Stephenson
 *   Date: Thu Oct  9 18:09:52 2003
 *
 *   Function: This library contains global variables that the loop
 *      instrumentor uses to instrument loops.  The functions 
 *      __measure_loop_init and __measure_loop_final initialize and
 *      finalize the library.  Other than that, the loop instrumentor
 *      inserts all the requisite code.  It should be noted however,
 *      that the inserted code accesses the global variables defined
 *      in this object.
 *
 ***************************************************************************/

#ifndef cg_loop_lib_INCLUDED
#define cg_loop_lib_INCLUDED

// ------------------------------------------------------------------    
// List of functions for measuring loop runtime.
void __measure_loop_init (char*, signed int);
void __measure_loop_final (char*, signed int);

// The loop timer array contains 2*N entries, where N is the number
// of loops that are instrumented.  If I is the id of a particular
// loop, the loop's running counter is the 2*Ith index, and the
// cumulative total is in 2*I+1.
extern unsigned long int *_loop_timers;

// This variable keeps track of the number of timers (i.e., the
// number of instrumented loops.
extern int _loop_timers_size;

// This variable is used as an error condition code register.
// If its value when __measure_loop_final is greater than or
// equal to NO_PROBLEMS, then there was a problem with the 
// associated loop.
extern int _loop_timers_error;

// This variable holds the name of the benchmarks you're running.
extern char *_loop_exec_name;

#endif

